==> edgedb-0.24.0a7-cp38-cp38-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: edgedb/options.py <==
import abc
import enum
import random
import typing
from collections import namedtuple

from . import errors


_RetryRule = namedtuple("_RetryRule", ["attempts", "backoff"])


def default_backoff(attempt):
    return (2 ** attempt) * 0.1 + random.randrange(100) * 0.001


class RetryCondition:
    """Specific condition to retry on for fine-grained control"""
    TransactionConflict = enum.auto()
    NetworkError = enum.auto()


class IsolationLevel:
    """Isolation level for transaction"""
    Serializable = "SERIALIZABLE"


class RetryOptions:
    """An immutable class that contains rules for `transaction()`"""
    __slots__ = ['_default', '_overrides']

    def __init__(self, attempts: int, backoff=default_backoff):
        self._default = _RetryRule(attempts, backoff)
        self._overrides = None

    def with_rule(self, condition, attempts=None, backoff=None):
        default = self._default
        overrides = self._overrides
        if overrides is None:
            overrides = {}
        else:
            overrides = overrides.copy()
        overrides[condition] = _RetryRule(
            default.attempts if attempts is None else attempts,
            default.backoff if backoff is None else backoff,
        )
        result = RetryOptions.__new__(RetryOptions)
        result._default = default
        result._overrides = overrides
        return result

    @classmethod
    def defaults(cls):
        return cls(
            attempts=3,
            backoff=default_backoff,
        )

    def get_rule_for_exception(self, exception):
        default = self._default
        overrides = self._overrides
        res = default
        if overrides:
            if isinstance(exception, errors.TransactionConflictError):
                res = overrides.get(RetryCondition.TransactionConflict, res)
            elif isinstance(exception, errors.ClientError):
                res = overrides.get(RetryCondition.NetworkError, res)
        return res


class TransactionOptions:
    """Options for `transaction()`"""
    __slots__ = ['_isolation', '_readonly', '_deferrable']

    def __init__(
        self,
        isolation: IsolationLevel=IsolationLevel.Serializable,
        readonly: bool = False,
        deferrable: bool = False,
    ):
        self._isolation = isolation
        self._readonly = readonly
        self._deferrable = deferrable

    @classmethod
    def defaults(cls):
        return cls()

    def start_transaction_query(self):
        isolation = str(self._isolation)
        if self._readonly:
            mode = 'READ ONLY'
        else:
            mode = 'READ WRITE'

        if self._deferrable:
            defer = 'DEFERRABLE'
        else:
            defer = 'NOT DEFERRABLE'

        return f'START TRANSACTION ISOLATION {isolation}, {mode}, {defer};'

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'isolation:{self._isolation}, '
            f'readonly:{self._readonly}, '
            f'deferrable:{self._deferrable}>'
        )


class State:
    __slots__ = ['_module', '_aliases', '_config', '_globals']

    def __init__(
        self,
        module: typing.Optional[str] = None,
        aliases: typing.Mapping[str, str] = None,
        config: typing.Mapping[str, typing.Any] = None,
        globals_: typing.Mapping[str, typing.Any] = None,
    ):
        self._module = module
        self._aliases = {} if aliases is None else dict(aliases)
        self._config = {} if config is None else dict(config)
        self._globals = {} if globals_ is None else dict(globals_)

    @classmethod
    def defaults(cls):
        return cls()

    def with_module_aliases(self, module=..., **aliases):
        new_aliases = self._aliases.copy()
        new_aliases.update(aliases)
        return State(
            module=self._module if module is ... else module,
            aliases=new_aliases,
            config=self._config,
            globals_=self._globals,
        )

    def with_config(self, **config):
        new_config = self._config.copy()
        new_config.update(config)
        return State(
            module=self._module,
            aliases=self._aliases,
            config=new_config,
            globals_=self._globals,
        )

    def with_globals(self, **globals_):
        new_globals = self._globals.copy()
        new_globals.update(globals_)
        return State(
            module=self._module,
            aliases=self._aliases,
            config=self._config,
            globals_=new_globals,
        )

    def as_dict(self):
        rv = {}
        if self._module is not None:
            module = rv["module"] = self._module
        else:
            module = 'default'
        if self._aliases:
            rv["aliases"] = list(self._aliases.items())
        if self._config:
            rv["config"] = self._config
        if self._globals:
            rv["globals"] = g = {}
            for k, v in self._globals.items():
                parts = k.split("::")
                if len(parts) == 1:
                    g[f"{module}::{k}"] = v
                elif len(parts) == 2:
                    mod, glob = parts
                    mod = self._aliases.get(mod, mod)
                    g[f"{mod}::{glob}"] = v
                else:
                    raise errors.InvalidArgumentError(
                        f"Illegal global name: {k}"
                    )
        return rv


class _OptionsMixin:
    def __init__(self, *args, **kwargs):
        self._options = _Options.defaults()
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def _shallow_clone(self):
        pass

    def with_transaction_options(self, options: TransactionOptions = None):
        """Returns object with adjusted options for future transactions.

        :param options TransactionOptions:
            Object that encapsulates transaction options.

        This method returns a "shallow copy" of the current object
        with modified transaction options.

        Both ``self`` and returned object can be used after, but when using
        them transaction options applied will be different.

        Transaction options are are used by the ``transaction`` method.
        """
        result = self._shallow_clone()
        result._options = self._options.with_transaction_options(options)
        return result

    def with_retry_options(self, options: RetryOptions=None):
        """Returns object with adjusted options for future retrying
        transactions.

        :param options RetryOptions:
            Object that encapsulates retry options.

        This method returns a "shallow copy" of the current object
        with modified retry options.

        Both ``self`` and returned object can be used after, but when using
        them transaction options applied will be different.
        """

        result = self._shallow_clone()
        result._options = self._options.with_retry_options(options)
        return result

    def with_state(self, state: State):
        result = self._shallow_clone()
        result._options = self._options.with_state(state)
        return result

    def with_module_aliases(self, module=None, **aliases):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_module_aliases(module=module, **aliases)
        )
        return result

    def with_config(self, **config):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_config(**config)
        )
        return result

    def with_globals(self, **globals_):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_globals(**globals_)
        )
        return result


class _Options:
    """Internal class for storing connection options"""

    __slots__ = ['_retry_options', '_transaction_options', '_state']

    def __init__(
        self,
        retry_options: RetryOptions,
        transaction_options: TransactionOptions,
        state: State,
    ):
        self._retry_options = retry_options
        self._transaction_options = transaction_options
        self._state = state

    @property
    def retry_options(self):
        return self._retry_options

    @property
    def transaction_options(self):
        return self._transaction_options

    @property
    def state(self):
        return self._state

    def with_retry_options(self, options: RetryOptions):
        return _Options(
            options,
            self._transaction_options,
            self._state,
        )

    def with_transaction_options(self, options: TransactionOptions):
        return _Options(
            self._retry_options,
            options,
            self._state,
        )

    def with_state(self, state: State):
        return _Options(
            self._retry_options,
            self._transaction_options,
            state,
        )

    @classmethod
    def defaults(cls):
        return cls(
            RetryOptions.defaults(),
            TransactionOptions.defaults(),
            State.defaults(),
        )

==> edgedb-1.2.0-cp38-cp38-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: edgedb/options.py <==
import abc
import enum
import random
import typing
from collections import namedtuple

from . import errors


_RetryRule = namedtuple("_RetryRule", ["attempts", "backoff"])


def default_backoff(attempt):
    return (2 ** attempt) * 0.1 + random.randrange(100) * 0.001


class RetryCondition:
    """Specific condition to retry on for fine-grained control"""
    TransactionConflict = enum.auto()
    NetworkError = enum.auto()


class IsolationLevel:
    """Isolation level for transaction"""
    Serializable = "SERIALIZABLE"


class RetryOptions:
    """An immutable class that contains rules for `transaction()`"""
    __slots__ = ['_default', '_overrides']

    def __init__(self, attempts: int, backoff=default_backoff):
        self._default = _RetryRule(attempts, backoff)
        self._overrides = None

    def with_rule(self, condition, attempts=None, backoff=None):
        default = self._default
        overrides = self._overrides
        if overrides is None:
            overrides = {}
        else:
            overrides = overrides.copy()
        overrides[condition] = _RetryRule(
            default.attempts if attempts is None else attempts,
            default.backoff if backoff is None else backoff,
        )
        result = RetryOptions.__new__(RetryOptions)
        result._default = default
        result._overrides = overrides
        return result

    @classmethod
    def defaults(cls):
        return cls(
            attempts=3,
            backoff=default_backoff,
        )

    def get_rule_for_exception(self, exception):
        default = self._default
        overrides = self._overrides
        res = default
        if overrides:
            if isinstance(exception, errors.TransactionConflictError):
                res = overrides.get(RetryCondition.TransactionConflict, res)
            elif isinstance(exception, errors.ClientError):
                res = overrides.get(RetryCondition.NetworkError, res)
        return res


class TransactionOptions:
    """Options for `transaction()`"""
    __slots__ = ['_isolation', '_readonly', '_deferrable']

    def __init__(
        self,
        isolation: IsolationLevel=IsolationLevel.Serializable,
        readonly: bool = False,
        deferrable: bool = False,
    ):
        self._isolation = isolation
        self._readonly = readonly
        self._deferrable = deferrable

    @classmethod
    def defaults(cls):
        return cls()

    def start_transaction_query(self):
        isolation = str(self._isolation)
        if self._readonly:
            mode = 'READ ONLY'
        else:
            mode = 'READ WRITE'

        if self._deferrable:
            defer = 'DEFERRABLE'
        else:
            defer = 'NOT DEFERRABLE'

        return f'START TRANSACTION ISOLATION {isolation}, {mode}, {defer};'

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'isolation:{self._isolation}, '
            f'readonly:{self._readonly}, '
            f'deferrable:{self._deferrable}>'
        )


class State:
    __slots__ = ['_module', '_aliases', '_config', '_globals']

    def __init__(
        self,
        default_module: typing.Optional[str] = None,
        module_aliases: typing.Mapping[str, str] = None,
        config: typing.Mapping[str, typing.Any] = None,
        globals_: typing.Mapping[str, typing.Any] = None,
    ):
        self._module = default_module
        self._aliases = {} if module_aliases is None else dict(module_aliases)
        self._config = {} if config is None else dict(config)
        self._globals = (
            {} if globals_ is None else self.with_globals(globals_)._globals
        )

    @classmethod
    def _new(cls, default_module, module_aliases, config, globals_):
        rv = cls.__new__(cls)
        rv._module = default_module
        rv._aliases = module_aliases
        rv._config = config
        rv._globals = globals_
        return rv

    @classmethod
    def defaults(cls):
        return cls()

    def with_default_module(self, module: typing.Optional[str] = None):
        return self._new(
            default_module=module,
            module_aliases=self._aliases,
            config=self._config,
            globals_=self._globals,
        )

    def with_module_aliases(self, *args, **aliases):
        if len(args) > 1:
            raise errors.InvalidArgumentError(
                "with_module_aliases() takes from 0 to 1 positional arguments "
                "but {} were given".format(len(args))
            )
        aliases_dict = args[0] if args else {}
        aliases_dict.update(aliases)
        new_aliases = self._aliases.copy()
        new_aliases.update(aliases_dict)
        return self._new(
            default_module=self._module,
            module_aliases=new_aliases,
            config=self._config,
            globals_=self._globals,
        )

    def with_config(self, *args, **config):
        if len(args) > 1:
            raise errors.InvalidArgumentError(
                "with_config() takes from 0 to 1 positional arguments "
                "but {} were given".format(len(args))
            )
        config_dict = args[0] if args else {}
        config_dict.update(config)
        new_config = self._config.copy()
        new_config.update(config_dict)
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=new_config,
            globals_=self._globals,
        )

    def resolve(self, name: str) -> str:
        parts = name.split("::")
        if len(parts) == 1:
            return f"{self._module or 'default'}::{name}"
        elif len(parts) == 2:
            mod, name = parts
            mod = self._aliases.get(mod, mod)
            return f"{mod}::{name}"
        else:
            raise errors.InvalidArgumentError(f"Illegal name: {name}")

    def with_globals(self, *args, **globals_):
        if len(args) > 1:
            raise errors.InvalidArgumentError(
                "with_globals() takes from 0 to 1 positional arguments "
                "but {} were given".format(len(args))
            )
        new_globals = self._globals.copy()
        if args:
            for k, v in args[0].items():
                new_globals[self.resolve(k)] = v
        for k, v in globals_.items():
            new_globals[self.resolve(k)] = v
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=self._config,
            globals_=new_globals,
        )

    def without_module_aliases(self, *aliases):
        if not aliases:
            new_aliases = {}
        else:
            new_aliases = self._aliases.copy()
            for alias in aliases:
                new_aliases.pop(alias, None)
        return self._new(
            default_module=self._module,
            module_aliases=new_aliases,
            config=self._config,
            globals_=self._globals,
        )

    def without_config(self, *config_names):
        if not config_names:
            new_config = {}
        else:
            new_config = self._config.copy()
            for name in config_names:
                new_config.pop(name, None)
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=new_config,
            globals_=self._globals,
        )

    def without_globals(self, *global_names):
        if not global_names:
            new_globals = {}
        else:
            new_globals = self._globals.copy()
            for name in global_names:
                new_globals.pop(self.resolve(name), None)
        return self._new(
            default_module=self._module,
            module_aliases=self._aliases,
            config=self._config,
            globals_=new_globals,
        )

    def as_dict(self):
        rv = {}
        if self._module is not None:
            rv["module"] = self._module
        if self._aliases:
            rv["aliases"] = list(self._aliases.items())
        if self._config:
            rv["config"] = self._config
        if self._globals:
            rv["globals"] = self._globals
        return rv


class _OptionsMixin:
    def __init__(self, *args, **kwargs):
        self._options = _Options.defaults()
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def _shallow_clone(self):
        pass

    def with_transaction_options(self, options: TransactionOptions = None):
        """Returns object with adjusted options for future transactions.

        :param options TransactionOptions:
            Object that encapsulates transaction options.

        This method returns a "shallow copy" of the current object
        with modified transaction options.

        Both ``self`` and returned object can be used after, but when using
        them transaction options applied will be different.

        Transaction options are used by the ``transaction`` method.
        """
        result = self._shallow_clone()
        result._options = self._options.with_transaction_options(options)
        return result

    def with_retry_options(self, options: RetryOptions=None):
        """Returns object with adjusted options for future retrying
        transactions.

        :param options RetryOptions:
            Object that encapsulates retry options.

        This method returns a "shallow copy" of the current object
        with modified retry options.

        Both ``self`` and returned object can be used after, but when using
        them retry options applied will be different.
        """

        result = self._shallow_clone()
        result._options = self._options.with_retry_options(options)
        return result

    def with_state(self, state: State):
        result = self._shallow_clone()
        result._options = self._options.with_state(state)
        return result

    def with_default_module(self, module: typing.Optional[str] = None):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_default_module(module)
        )
        return result

    def with_module_aliases(self, *args, **aliases):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_module_aliases(*args, **aliases)
        )
        return result

    def with_config(self, *args, **config):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_config(*args, **config)
        )
        return result

    def with_globals(self, *args, **globals_):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.with_globals(*args, **globals_)
        )
        return result

    def without_module_aliases(self, *aliases):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.without_module_aliases(*aliases)
        )
        return result

    def without_config(self, *config_names):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.without_config(*config_names)
        )
        return result

    def without_globals(self, *global_names):
        result = self._shallow_clone()
        result._options = self._options.with_state(
            self._options.state.without_globals(*global_names)
        )
        return result


class _Options:
    """Internal class for storing connection options"""

    __slots__ = ['_retry_options', '_transaction_options', '_state']

    def __init__(
        self,
        retry_options: RetryOptions,
        transaction_options: TransactionOptions,
        state: State,
    ):
        self._retry_options = retry_options
        self._transaction_options = transaction_options
        self._state = state

    @property
    def retry_options(self):
        return self._retry_options

    @property
    def transaction_options(self):
        return self._transaction_options

    @property
    def state(self):
        return self._state

    def with_retry_options(self, options: RetryOptions):
        return _Options(
            options,
            self._transaction_options,
            self._state,
        )

    def with_transaction_options(self, options: TransactionOptions):
        return _Options(
            self._retry_options,
            options,
            self._state,
        )

    def with_state(self, state: State):
        return _Options(
            self._retry_options,
            self._transaction_options,
            state,
        )

    @classmethod
    def defaults(cls):
        return cls(
            RetryOptions.defaults(),
            TransactionOptions.defaults(),
            State.defaults(),
        )
